#!/usr/bin/env python
# encoding: utf-8

"""
File       : XMLUtils.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Set of utilities for RequestManager code
"""

from __future__ import (division, print_function)
from builtins import str, bytes
from Utils.Utilities import encodeUnicodeToBytes

from io import BytesIO
import re
import xml.etree.cElementTree as ET

int_number_pattern = re.compile(r'(^[0-9-]$|^[0-9-][0-9]*$)')
float_number_pattern = re.compile(r'(^[-]?\d+\.\d*$|^\d*\.{1,1}\d+$)')


def adjust_value(value):
    """
    Change null value to None.
    """
    pat_float = float_number_pattern
    pat_integer = int_number_pattern
    if isinstance(value, str):
        if value == 'null' or value == '(null)':
            return None
        elif pat_float.match(value):
            return float(value)
        elif pat_integer.match(value):
            return int(value)
        else:
            return value
    else:
        return value


def xml_parser(data, prim_key):
    """
    Generic XML parser

    :param data: can be of type "file object", unicode string or bytes string
    """
    if isinstance(data, (str, bytes)):
        stream = BytesIO()
        data = encodeUnicodeToBytes(data, "ignore")
        stream.write(data)
        stream.seek(0)
    else:
        stream = data

    context = ET.iterparse(stream)
    for event, elem in context:
        row = {}
        key = elem.tag
        if key != prim_key:
            continue
        row[key] = elem.attrib
        get_children(elem, event, row, key)
        elem.clear()
        yield row


def get_children(elem, event, row, key):
    """
    xml_parser helper function. It gets recursively information about
    children for given element tag. Information is stored into provided
    row for given key. The change of notations can be applied during
    parsing step by using provided notations dictionary.
    """
    for child in list(elem):
        child_key = child.tag
        child_data = child.attrib
        if not child_data:
            child_dict = adjust_value(child.text)
        else:
            child_dict = child_data

        if list(child):  # we got grand-children
            if child_dict:
                row[key][child_key] = child_dict
            else:
                row[key][child_key] = {}
            if isinstance(child_dict, dict):
                newdict = {child_key: child_dict}
            else:
                newdict = {child_key: {}}
            get_children(child, event, newdict, child_key)
            row[key][child_key] = newdict[child_key]
        else:
            if not isinstance(row[key], dict):
                row[key] = {}
            row[key].setdefault(child_key, [])
            row[key][child_key].append(child_dict)
        child.clear()



"""
File       : MSRuleCleaner.py

Description:
This MicroService is meant to remove Rucio rules that are no longer needed in the Workload Management system, such as:
 * block-level     rules created by WMAgent, against the origin RSE where data is getting produced
 * container-level rules created by WMAgent, with a very generic RSE expression
 * block-level     rules created by MSTransferor, for input data that is no longer in the system
 * container-level rules created by MSTransferor, for input data that is no longer in the system
In addition to that, this MicroService is now also responsible for the workflow archival, which is the final status that
workflows remain.
"""
# TODO: To Add the top level description of the logic into the module's Docstring

# futures
from __future__ import division, print_function
from future.utils import viewvalues

import json
import re
import time

# system modules
from threading import current_thread
from pprint import pformat

# WMCore modules
from WMCore.MicroService.DataStructs.DefaultStructs import RULECLEANER_REPORT
from WMCore.MicroService.MSRuleCleaner.MSRuleCleanerWflow import MSRuleCleanerWflow
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.Services.pycurl_manager import RequestHandler
from WMCore.Services.Rucio.Rucio import WMRucioDIDNotFoundException
from WMCore.ReqMgr.DataStructs import RequestStatus
from WMCore.WMException import WMException
from WMCore.Services.LogDB.LogDB import LogDB
from WMCore.Services.WMStatsServer.WMStatsServer import WMStatsServer
from WMCore.MicroService.Tools.Common import findParent
from Utils.Pipeline import Pipeline, Functor
from Utils.CertTools import ckey, cert


class MSRuleCleanerResolveParentError(WMException):
    """
    WMCore exception class for the MSRuleCleaner module, in WMCore MicroServices,
    used to signal if an error occurred during parent dataset resolution step.
    """
    def __init__(self, message):
        super(MSRuleCleanerResolveParentError, self).__init__(message)


class MSRuleCleanerArchivalError(WMException):
    """
    Archival Exception Class for MSRuleCleaner Module in WMCore MicroServices
    used to signal an abnormal condition if occurred during the archival step.
    """
    def __init__(self, message):
        super(MSRuleCleanerArchivalError, self).__init__(message)


class MSRuleCleanerArchivalSkip(WMException):
    """
    Archival Exception Class for MSRuleCleaner Module in WMCore MicroServices
    used to signal an expected condition which should interrupt the archival process
    """
    def __init__(self, message):
        super(MSRuleCleanerArchivalSkip, self).__init__(message)


class MSRuleCleaner(MSCore):
    """
    MSRuleCleaner.py class provides the logic used to clean the Rucio
    block level data placement rules created by WMAgent.
    """

    def __init__(self, msConfig, logger=None):
        """
        Runs the basic setup and initialization for the MSRuleCleaner module
        :param msConfig: micro service configuration
        """
        super(MSRuleCleaner, self).__init__(msConfig, logger=logger)

        self.msConfig.setdefault("verbose", True)
        self.msConfig.setdefault("interval", 60)
        self.msConfig.setdefault("services", ['ruleCleaner'])
        self.msConfig.setdefault("rucioWmaAccount", "wma_test")
        self.msConfig.setdefault("rucioMStrAccount", "wmcore_transferor")
        self.msConfig.setdefault('enableRealMode', False)
        self.msConfig.setdefault('archiveDelayHours', 24 * 2)
        self.msConfig.setdefault('archiveAlarmHours', 24 * 30)
        self.msConfig.setdefault("sendNotification", False)

        self.currThread = None
        self.currThreadIdent = None

        self.mode = "RealMode" if self.msConfig['enableRealMode'] else "DryRunMode"
        self.curlMgr = RequestHandler()
        self.targetStatusRegex = re.compile(r'.*archived')
        self.logDB = LogDB(self.msConfig["logDBUrl"],
                           self.msConfig["logDBReporter"],
                           logger=self.logger)
        self.wmstatsSvc = WMStatsServer(self.msConfig['wmstatsUrl'], logger=self.logger)
        # service name used to route alerts via AlertManager
        self.alertServiceName = "ms-rulecleaner"
        # 2 days of expiration time
        self.alertExpiration = self.msConfig.get("alertExpireSecs", 2 * 24 * 60 * 60)

        # Building all the Pipelines:
        pName = 'plineMSTrCont'
        self.plineMSTrCont = Pipeline(name=pName,
                                      funcLine=[Functor(self.setPlineMarker, pName),
                                                Functor(self.setParentDatasets),
                                                Functor(self.getRucioRules, 'container', self.msConfig['rucioMStrAccount']),
                                                Functor(self.cleanRucioRules)])
        pName = 'plineMSTrBlock'
        self.plineMSTrBlock = Pipeline(name=pName,
                                       funcLine=[Functor(self.setPlineMarker, pName),
                                                 Functor(self.setParentDatasets),
                                                 Functor(self.getRucioRules, 'block', self.msConfig['rucioMStrAccount']),
                                                 Functor(self.cleanRucioRules)])
        pName = 'plineAgentCont'
        self.plineAgentCont = Pipeline(name=pName,
                                       funcLine=[Functor(self.setPlineMarker, pName),
                                                 Functor(self.getRucioRules, 'container', self.msConfig['rucioWmaAccount']),
                                                 Functor(self.cleanRucioRules)])
        pName = 'plineAgentBlock'
        self.plineAgentBlock = Pipeline(name=pName,
                                        funcLine=[Functor(self.setPlineMarker, pName),
                                                  Functor(self.getRucioRules, 'block', self.msConfig['rucioWmaAccount']),
                                                  Functor(self.cleanRucioRules)])
        pName = 'plineArchive'
        self.plineArchive = Pipeline(name=pName,
                                     funcLine=[Functor(self.setPlineMarker, pName),
                                               Functor(self.findTargetStatus),
                                               Functor(self.setClean),
                                               Functor(self.setArchivalDelayExpired),
                                               Functor(self.setLogDBClean),
                                               Functor(self.archive)])

        # Building the different set of plines we will need later:
        # NOTE: The following are all the functional pipelines which are supposed to include
        #       a cleanup function and report cleanup status in the MSRuleCleanerWflow object
        self.cleanuplines = [self.plineMSTrCont,
                             self.plineMSTrBlock,
                             self.plineAgentCont,
                             self.plineAgentBlock]
        # Building an auxiliary list of cleanup pipeline names only:
        self.cleanupPipeNames = [pline.name for pline in self.cleanuplines]

        # Building lists of pipelines related only to Agents or MStransferror
        self.agentlines = [self.plineAgentCont,
                           self.plineAgentBlock]
        self.mstrlines = [self.plineMSTrCont,
                          self.plineMSTrBlock]

        # Initialization of the 'cleaned' and 'archived' counters:
        self.wfCounters = {'cleaned': {},
                           'archived': {'normalArchived': 0,
                                        'forceArchived': 0}}
        self.globalLocks = set()

    def getGlobalLocks(self):
        """
        Fetches the list of 'globalLocks' from wmstats server and the list of
        'parentLocks' from request manager. Stores/updates the unified set in
        the 'globalLocks' instance variable. Returns the resultant unified set.
        :return: A union set of the 'globalLocks' and the 'parentLocks' lists
        """
        self.logger.info("Fetching globalLocks list from wmstats server.")
        try:
            globalLocks = set(self.wmstatsSvc.getGlobalLocks())
        except Exception as ex:
            msg = "Failed to refresh global locks list for the current polling cycle. Error: %s "
            msg += "Skipping this polling cycle."
            self.logger.error(msg, str(ex))
            raise ex
        self.logger.info("Fetching parentLocks list from reqmgr2 server.")
        try:
            parentLocks = set(self.reqmgr2.getParentLocks())
        except Exception as ex:
            msg = "Failed to refresh parent locks list for the current poling cycle. Error: %s "
            msg += "Skipping this polling cycle."
            self.logger.error(msg, str(ex))
            raise ex
        self.globalLocks = globalLocks | parentLocks

    def resetCounters(self):
        """
        A simple function for zeroing the cleaned and archived counters.
        """
        for pline in self.cleanuplines:
            self.wfCounters['cleaned'][pline.name] = 0
        self.wfCounters['archived']['normalArchived'] = 0
        self.wfCounters['archived']['forceArchived'] = 0

    def execute(self, reqStatus):
        """
        Executes the whole ruleCleaner logic
        :return: summary
        """
        # start threads in MSManager which should call this method
        summary = dict(RULECLEANER_REPORT)

        self.currThread = current_thread()
        self.currThreadIdent = self.currThread.name
        self.updateReportDict(summary, "thread_id", self.currThreadIdent)
        self.resetCounters()
        self.logger.info("MSRuleCleaner is running in mode: %s.", self.mode)

        # Build the list of workflows to work on:
        try:
            requestRecords = {}
            for status in reqStatus:
                requestRecords.update(self.getRequestRecords(status))
        except Exception as err:  # general error
            msg = "Unknown exception while fetching requests from ReqMgr2. Error: %s", str(err)
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

        # Call _execute() and feed the relevant pipeline with the objects popped from requestRecords
        try:
            self.getGlobalLocks()
            totalNumRequests, cleanNumRequests, normalArchivedNumRequests, forceArchivedNumRequests = self._execute(requestRecords)
            msg = "\nNumber of processed workflows: %s."
            msg += "\nNumber of properly cleaned workflows: %s."
            msg += "\nNumber of normally archived workflows: %s."
            msg += "\nNumber of force archived workflows: %s."
            self.logger.info(msg,
                             totalNumRequests,
                             cleanNumRequests,
                             normalArchivedNumRequests,
                             forceArchivedNumRequests)
            self.updateReportDict(summary, "total_num_requests", totalNumRequests)
            self.updateReportDict(summary, "clean_num_requests", cleanNumRequests)
            self.updateReportDict(summary, "normal_archived_num_requests", normalArchivedNumRequests)
            self.updateReportDict(summary, "force_archived_num_requests", forceArchivedNumRequests)
        except Exception as ex:
            msg = "Unknown exception while running MSRuleCleaner thread Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

        return summary

    def _execute(self, reqRecords):
        """
        Executes the MSRuleCleaner pipelines based on the workflow status
        :param reqList: A list of RequestRecords to work on
        :return:        a tuple with:
                            number of properly cleaned requests
                            number of processed workflows
                            number of archived workflows
                            number of forced archived workflows
        """
        # NOTE: The Input Cleanup, the Block Level Cleanup and the Archival
        #       Pipelines are executed sequentially in the above order.
        #       This way we assure ourselves that we archive only workflows
        #       that have accomplished the needed cleanup

        cleanNumRequests = 0
        totalNumRequests = 0

        # Call the workflow dispatcher:
        for req in viewvalues(reqRecords):
            wflow = MSRuleCleanerWflow(req)
            self._dispatchWflow(wflow)
            msg = "\n----------------------------------------------------------"
            msg += "\nMSRuleCleanerWflow: %s"
            msg += "\n----------------------------------------------------------"
            self.logger.debug(msg, pformat(wflow))
            totalNumRequests += 1
            if self._checkClean(wflow):
                cleanNumRequests += 1

        # Report the counters:
        for pline in self.cleanuplines:
            msg = "Workflows cleaned by pipeline: %s: %d"
            self.logger.info(msg, pline.name, self.wfCounters['cleaned'][pline.name])
        normalArchivedNumRequests = self.wfCounters['archived']['normalArchived']
        forceArchivedNumRequests = self.wfCounters['archived']['forceArchived']
        self.logger.info("Workflows normally archived: %d", self.wfCounters['archived']['normalArchived'])
        self.logger.info("Workflows force archived: %d", self.wfCounters['archived']['forceArchived'])
        return totalNumRequests, cleanNumRequests, normalArchivedNumRequests, forceArchivedNumRequests

    def _dispatchWflow(self, wflow):
        """
        A function intended to dispatch a workflow (e.g based on its status)
        through one or more functional pipelines in case there is some more
        complicated logic involved in the order we execute them but not just
        a sequentially
        """
        self.logger.debug("Dispatching workflow: %s", wflow['RequestName'])
        # NOTE: The following dispatch logic is a subject to be changed at any time

        # Resolve:
        # NOTE: First resolve any preliminary flags that will be needed further
        #       in the logic of the _dispatcher() itself
        if wflow['RequestStatus'] == 'announced':
            self.getMSOutputTransferInfo(wflow)

        # Clean:
        # Do not clean any Resubmission, but still let them be archived
        if wflow['RequestType'] == 'Resubmission':
            wflow['ForceArchive'] = True
            msg = "Skipping cleanup step for workflow: %s - RequestType is %s."
            msg += " Will try to archive it directly."
            self.logger.info(msg, wflow['RequestName'], wflow['RequestType'])
        elif wflow['RequestStatus'] == 'announced' and not wflow['TransferDone']:
            # NOTE: We skip workflows which have not yet finalised their TransferStatus
            #       in MSOutput, but we still need some proper logging for them.
            msg = "Skipping workflow: %s - 'TransferStatus' is 'pending' or 'TransferInfo' is missing in MSOutput." % wflow['RequestName']
            msg += " Will retry again in the next cycle."
            self.logger.info(msg)
            self._checkStatusAdvanceExpired(wflow, additionalInfo=msg)
        elif wflow['RequestStatus'] == 'announced' and not wflow['TransferTape']:
            # Given that we are still waiting for the tape transfers to be fulfilled,
            # we can go ahead and start cleaning up the input data.
            msg = "Skipping workflow: %s - tape transfers are not yet completed." % wflow['RequestName']
            msg += "Workflow in 'announced' state, hence proceeding only with MSTransferor / input data removal. "  
            msg += "Will retry the remaining in the next cycle."
            self.logger.info(msg)
            self._checkStatusAdvanceExpired(wflow, additionalInfo=msg)
            for pline in self.mstrlines:
                try: 
                    pline.run(wflow)
                except Exception as ex:
                    msg = f"{pline.name}: General error from pipeline"
                    msg += " when cleaning input MSTransferor rules."
                    msg += f"\nWorkflow: {wflow['RequestName']}. Error:  \n{str(ex)}."
                    msg += "\nWill retry again in the next cycle."
                    self.logger.exception(msg)
                    continue
            # return now to avoid exeecuting the archival pipeline
            return
        elif wflow['RequestStatus'] in ['announced', 'rejected', 'aborted-completed']:
            # Workflows reaching this block are ready for the full pipeline execution
            for pline in self.cleanuplines:
                try:
                    pline.run(wflow)
                except MSRuleCleanerResolveParentError as ex:
                    msg = "%s: Parentage Resolve Error: %s. " % (pline.name, str(ex))
                    msg += "Will retry again in the next cycle."
                    self.logger.error(msg)
                    self._checkStatusAdvanceExpired(wflow, additionalInfo=msg)
                    continue
                except Exception as ex:
                    msg = "%s: General error from pipeline. Workflow: %s. Error:  \n%s. " % (pline.name, wflow['RequestName'], str(ex))
                    msg += "\nWill retry again in the next cycle."
                    self.logger.exception(msg)
                    self._checkStatusAdvanceExpired(wflow, additionalInfo=msg)
                    continue
                if wflow['CleanupStatus'][pline.name]:
                    self.wfCounters['cleaned'][pline.name] += 1
        else:
            # We shouldn't be here:
            msg = "Skipping workflow: %s - "
            msg += "Does not fall under any of the defined categories."
            self.logger.error(msg, wflow['RequestName'])

        # Archive:
        try:
            self.plineArchive.run(wflow)
            if wflow['ForceArchive']:
                self.wfCounters['archived']['forceArchived'] += 1
            else:
                self.wfCounters['archived']['normalArchived'] += 1
        except MSRuleCleanerArchivalSkip as ex:
            msg = "%s: Proper conditions not met: %s. "
            msg += "Skipping archival in the current cycle."
            self.logger.info(msg, wflow['PlineMarkers'][-1], str(ex))
        except MSRuleCleanerArchivalError as ex:
            msg = "%s: Archival Error: %s. "
            msg += "Will retry again in the next cycle."
            self.logger.error(msg, wflow['PlineMarkers'][-1], str(ex))
        except Exception as ex:
            msg = "%s General error from pipeline. Workflow: %s. Error: \n%s. "
            msg += "\nWill retry again in the next cycle."
            self.logger.exception(msg, wflow['PlineMarkers'][-1], wflow['RequestName'], str(ex))

    def setPlineMarker(self, wflow, pName):
        """
        A function intended to mark which is the pipeline currently working
        on the workflow. It is supposed to be called always as a first function
        in the pipeline.
        :param  wflow:   A MSRuleCleaner workflow representation
        :param  pName:   The name of the functional pipeline
        :return:         The workflow object
        """
        # NOTE: The current functional pipeline MUST always be appended at the
        #       end of the 'PlineMarkers' list

        # First get rid of the default:
        if not wflow['PlineMarkers']:
            wflow['PlineMarkers'] = []

        # Then push our current value into the markers list:
        wflow['PlineMarkers'].append(pName)

        # Populate the list of flags to be used later:
        if pName not in wflow['RulesToClean']:
            if pName in self.cleanupPipeNames:
                wflow['RulesToClean'][pName] = []
        if pName not in wflow['CleanupStatus']:
            if pName in self.cleanupPipeNames:
                wflow['CleanupStatus'][pName] = False
        return wflow

    def _getLastStatusTransitionTime(self, wflow):
        """
        A method to return status transition time for a workflow.
        :param wflow:  A MSRuleCleanerWorkflow instance
        :return:       The UTC timestamp for the transition if it exists, None otherwise
        """
        try:
            for trans in wflow['RequestTransition'][::-1]:
                if trans['Status'] == wflow['RequestStatus']:
                    return trans['UpdateTime']
        except KeyError:
            self.logger.error("Missing or broken status transition history for %s", wflow['RequestName'])
            return None

    def _checkStatusAdvanceExpired(self, wflow, additionalInfo=""):
        """
        A method to check if a given status transition has timed out e.g. before
        an alarm being set. The timeout should be configurable and hence taken
        from the instance attributes. Preserve any additional info inside the wflow object.
        :param wflow:          A MSRuleCleanerWorkflow instance
        :param additionalInfo: A string with additional info to be preserved within the workflow for later use by the alarm itself. (Default: "")
        :return:               Bool: True if status alarm time has expired False otherwise
        """
        statusAdvanceExpired = False
        currentTime = int(time.time())
        alarmThreshold = self.msConfig['archiveAlarmHours'] * 3600
        transitionTime = self._getLastStatusTransitionTime(wflow)

        if transitionTime and (currentTime - transitionTime) > alarmThreshold:
            statusAdvanceExpired = True
            if wflow['StatusAdvanceExpiredMsg']:
                wflow['StatusAdvanceExpiredMsg'] += "\n"
            wflow['StatusAdvanceExpiredMsg'] += additionalInfo
        return statusAdvanceExpired

    def _checkClean(self, wflow):
        """
        An auxiliary function used to only check the temporary cleanup status.
        It basically takes the pipelines registered in 'PlineMarkers' that have
        already worked on the workflow as a mask and applies this mask over
        the set of flags in the 'CleanupStatus' field and then reduces the
        result to a single bool value
        """
        # NOTE: This is one of the few functions taking a workflow as an argument
        #       but returning a bool, since it is an auxiliary function and is not
        #       supposed to be called as a standalone function in a pipeline.
        # NOTE: `all([]) == True`, ergo all the 'rejected' && 'aborted-completed' workflows
        #       are also counted as properly cleaned and can trigger archival later

        # Build a list of bool flags based on the mask of PlineMarkers
        cleanFlagsList = [wflow['CleanupStatus'][key]
                          for key in wflow['PlineMarkers']
                          if key in wflow['CleanupStatus']]

        # If no one have worked on the workflow set the clean status to false
        if not wflow['PlineMarkers']:
            cleanStatus = False
        # If we have a mask longer than the list of flags avoid false positives
        # because of the behavior explained above - `all([]) == True`
        elif not cleanFlagsList:
            cleanStatus = False
        # Figure out the final value
        else:
            cleanStatus = all(cleanFlagsList)
        return cleanStatus

    def setClean(self, wflow):
        """
        A function to set the 'IsClean' flag based on the status from all the
        pipelines which have worked on the workflow (and have put their markers
        in the 'PlineMarkers' list)
        :param  wflow:      A MSRuleCleaner workflow representation
        :return:            The workflow object
        """
        wflow['IsClean'] = self._checkClean(wflow)
        return wflow

    def _checkLogDBClean(self, wflow):
        """
        An auxiliary function used to only check the LogDB cleanup status.
        It makes a query to LogDB in order to verify there are no any records for
        the current workflow
        :param wflow:       A MSRuleCleaner workflow representation
        :return:            True if no records were found in LogDB about wflow
        """
        cleanStatus = False
        logDBRecords = self.logDB.get(wflow['RequestName'])
        self.logger.debug("logDBRecords: %s", pformat(logDBRecords))
        if not logDBRecords:
            cleanStatus = True
        return cleanStatus

    def setLogDBClean(self, wflow):
        """
        A function to set the 'IsLogDBClean' flag based on the presence of any
        records in LogDB for the current workflow.
        :param  wflow:      A MSRuleCleaner workflow representation
        :return:            The workflow object
        """
        wflow['IsLogDBClean'] = self._checkLogDBClean(wflow)
        if not wflow['IsLogDBClean'] and wflow['IsArchivalDelayExpired']:
            wflow['IsLogDBClean'] = self._cleanLogDB(wflow)
        return wflow

    def _cleanLogDB(self, wflow):
        """
        A function to be used for cleaning all the records related to a workflow in logDB.
        :param wflow:       A MSRuleCleaner workflow representation
        :return:            True if NO errors were encountered while deleting
                            records from LogDB
        """
        cleanStatus = False
        try:
            if self.msConfig['enableRealMode']:
                self.logger.info("Deleting %s records from LogDB WMStats...", wflow['RequestName'])
                res = self.logDB.delete(wflow['RequestName'], agent=False)
                if res == 'delete-error':
                    msg = "Failed to delete logDB docs for wflow: %s" % wflow['RequestName']
                    raise MSRuleCleanerArchivalError(msg)
                cleanStatus = True
            else:
                self.logger.info("DRY-RUN: NOT Deleting %s records from LogDB WMStats...", wflow['RequestName'])
        except Exception as ex:
            msg = "General Exception while cleaning LogDB records for wflow: %s : %s"
            self.logger.exception(msg, wflow['RequestName'], str(ex))
        return cleanStatus

    def findTargetStatus(self, wflow):
        """
        Find the proper targeted archival status
        :param  wflow:      A MSRuleCleaner workflow representation
        :return:            The workflow object
        """
        # Check the available status transitions before we decide the final status
        targetStatusList = RequestStatus.REQUEST_STATE_TRANSITION.get(wflow['RequestStatus'], [])
        for status in targetStatusList:
            if self.targetStatusRegex.match(status):
                wflow['TargetStatus'] = status
        self.logger.debug("TargetStatus: %s", wflow['TargetStatus'])
        return wflow

    def _checkArchDelayExpired(self, wflow):
        """
        A function to check Archival Expiration Delay based on the information
        returned by WMStatsServer regarding the time of the last request status transition
        :param wflow:      MSRuleCleaner workflow representation
        :return:           True if the archival delay have been expired
        """
        archDelayExpired = False
        currentTime = int(time.time())
        threshold = self.msConfig['archiveDelayHours'] * 3600
        lastTransitionTime = self._getLastStatusTransitionTime(wflow)
        if lastTransitionTime and (currentTime - lastTransitionTime) > threshold:
            archDelayExpired = True
        return archDelayExpired

    def setArchivalDelayExpired(self, wflow):
        """
        A function to set the 'IsArchivalDelayExpired' flag
        """
        wflow['IsArchivalDelayExpired'] = self._checkArchDelayExpired(wflow)
        return wflow

    def archive(self, wflow):
        """
        Move the workflow to the proper archived status after checking
        the full cleanup status
        :param  wflow:      A MSRuleCleaner workflow representation
        :return:            The workflow object
        """
        # Make all the needed checks before trying to archive
        if not (wflow['ParentageResolved'] or wflow['ForceArchive']):
            msg = f"Not properly cleaned workflow: {wflow['RequestName']} - 'ParentageResolved' flag set to false."
            if self._checkStatusAdvanceExpired(wflow, additionalInfo=msg):
                self.alertStatusAdvanceExpired(wflow)
        if not (wflow['IsClean'] or wflow['ForceArchive']):
            msg = "Not properly cleaned workflow: %s" % wflow['RequestName']
            if self._checkStatusAdvanceExpired(wflow, additionalInfo=msg):
                self.alertStatusAdvanceExpired(wflow)
            raise MSRuleCleanerArchivalSkip(msg)
        if not wflow['TargetStatus']:
            msg = "Could not determine which archival status to target for workflow: %s" % wflow['RequestName']
            if self._checkStatusAdvanceExpired(wflow, additionalInfo=msg):
                self.alertStatusAdvanceExpired(wflow)
            raise MSRuleCleanerArchivalError(msg)
        if not wflow['IsLogDBClean']:
            msg = "LogDB records have not been cleaned for workflow: %s" % wflow['RequestName']
            if self._checkStatusAdvanceExpired(wflow, additionalInfo=msg):
                self.alertStatusAdvanceExpired(wflow)
            raise MSRuleCleanerArchivalSkip(msg)
        if not wflow['IsArchivalDelayExpired']:
            msg = "Archival delay period has not yet expired for workflow: %s." % wflow['RequestName']
            raise MSRuleCleanerArchivalSkip(msg)
        if not self.msConfig['enableRealMode']:
            msg = "Real Run Mode not enabled."
            raise MSRuleCleanerArchivalSkip(msg)

        # Proceed with the actual archival:
        try:
            self.reqmgr2.updateRequestStatus(wflow['RequestName'], wflow['TargetStatus'])
            msg = "Successful status transition to: %s for workflow: %s"
            self.logger.info(msg, wflow['TargetStatus'], wflow['RequestName'])
        except Exception as ex:
            msg = "General Exception while trying status transition to: %s " % wflow['TargetStatus']
            msg += "for workflow: %s : %s" % (wflow['RequestName'], str(ex))
            raise MSRuleCleanerArchivalError(msg) from None
        return wflow

    def getMSOutputTransferInfo(self, wflow):
        """
        Fetches the transfer information from the MSOutput REST interface for
        the given workflow.
        :param  wflow:   A MSRuleCleaner workflow representation
        :return:         The workflow object
        """
        headers = {'Accept': 'application/json'}
        params = {}
        url = '%s/data/info?request=%s' % (self.msConfig['msOutputUrl'],
                                           wflow['RequestName'])
        try:
            res = self.curlMgr.getdata(url, params=params, headers=headers, ckey=ckey(), cert=cert())
            data = json.loads(res)['result'][0]
            transferInfo = data['transferDoc']
        except Exception as ex:
            transferInfo = None
            msg = "General exception while fetching TransferInfo from MSOutput for %s. "
            msg += "Error: %s"
            self.logger.exception(msg, wflow['RequestName'], str(ex))

        if transferInfo is None:
            msg = f"Workflow {wflow['RequestName']} is still missing the output transfer document."
            self.logger.warning(msg)
        elif transferInfo['TransferStatus'] == 'pending':
            msg = f"Workflow {wflow['RequestName']} is still pending the output data placement."
            self.logger.warning(msg)
        elif transferInfo['TransferStatus'] == 'done':
            # Set Transfer status - information fetched from MSOutput only
            wflow['TransferDone'] = True

            # Set Tape rules status - information fetched from Rucio (tape rule ids from MSOutput)
            # For setting 'TransferTape' = True we require either no tape rules for the
            # workflow have been created or all existing tape rules to be in status 'OK',
            # so every empty TapeRuleID we consider as completed.
            tapeRulesStatusList = []
            for mapRecord in transferInfo['OutputMap']:
                if not mapRecord['TapeRuleID']:
                    continue

                rucioRule = self.rucio.getRule(mapRecord['TapeRuleID'])
                if not rucioRule:
                    msg = "Tape rule: %s not found for workflow: %s "
                    msg += "Possible server side error."
                    self.logger.error(msg, mapRecord['TapeRuleID'], wflow['RequestName'])
                    rucioRule = {'state': 'Missing'}

                if rucioRule['state'] == 'OK':
                    tapeRulesStatusList.append(True)
                    msg = "Tape rule: %s in final state: %s for workflow: %s"
                    self.logger.info(msg, mapRecord['TapeRuleID'], rucioRule['state'], wflow['RequestName'])
                else:
                    tapeRulesStatusList.append(False)
                    msg = "Tape rule: %s in non final state: %s for workflow: %s"
                    self.logger.info(msg, mapRecord['TapeRuleID'], rucioRule['state'], wflow['RequestName'])

                wflow['TapeRulesStatus'].append((mapRecord['TapeRuleID'], rucioRule['state'], mapRecord['Dataset']))

            if all(tapeRulesStatusList):
                wflow['TransferTape'] = True

        return wflow

    def setParentDatasets(self, wflow):
        """
        Used to resolve parent datasets for a workflow.
        :param  wflow:   A MSRuleCleaner workflow representation
        :return:         The workflow object
        """
        if wflow['InputDataset'] and wflow['IncludeParents']:
            childDataset = wflow['InputDataset']
            parentDataset = findParent([childDataset], self.msConfig['dbsUrl'])
            # NOTE: If findParent() returned None then the DBS service failed to
            #       resolve the request (it is considered an ERROR outside WMCore)
            if parentDataset.get(childDataset, None) is None:
                msg = "Failed to resolve parent dataset for: %s in workflow: %s" % (childDataset, wflow['RequestName'])
                raise MSRuleCleanerResolveParentError(msg)
            elif parentDataset:
                wflow['ParentDataset'] = [parentDataset[childDataset]]
                msg = "Found parent %s for input dataset %s in workflow: %s "
                self.logger.info(msg, parentDataset, wflow['InputDataset'], wflow['RequestName'])
            else:
                msg = "Could not find parent for input dataset: %s in workflows: %s"
                self.logger.error(msg, wflow['InputDataset'], wflow['RequestName'])
        return wflow

    def getRucioRules(self, wflow, gran, rucioAcct):
        """
        Queries Rucio and builds the relevant list of blocklevel rules for
        the given workflow
        :param  wflow:   A MSRuleCleaner workflow representation
        :param  gran:    Data granularity to search for Rucio rules. Possible values:
                         'block' or 'container'
        :return:         The workflow object
        """
        currPline = wflow['PlineMarkers'][-1]

        # Create the container list to the rucio account map and set the checkGlobalLocks flag.
        mapRuleType = {self.msConfig['rucioWmaAccount']: ["OutputDatasets"],
                       self.msConfig['rucioMStrAccount']: ["InputDataset", "ParentDataset"]}
        if rucioAcct == self.msConfig['rucioMStrAccount']:
            checkGlobalLocks = True
        else:
            checkGlobalLocks = False

        # Find all the data placement rules created by the components:
        for dataType in mapRuleType[rucioAcct]:
            dataList = wflow[dataType] if isinstance(wflow[dataType], list) else [wflow[dataType]]
            for dataCont in dataList:
                if dataCont is None:
                    continue
                self.logger.debug("getRucioRules: dataCont: %s", pformat(dataCont))
                if checkGlobalLocks and dataCont in self.globalLocks:
                    msg = "Found dataset: %s in GlobalLocks. NOT considering it for filling the "
                    msg += "RulesToClean list for both container and block level Rules for workflow: %s!"
                    self.logger.info(msg, dataCont, wflow['RequestName'])
                    continue
                if gran == 'container':
                    ruleIds = [rule['id'] for rule in self.rucio.listDataRules(dataCont, account=rucioAcct)]
                    if ruleIds:
                        wflow['RulesToClean'][currPline].extend(ruleIds)
                        msg = "Container %s has the following container-level rules to be removed: %s"
                        self.logger.info(msg, dataCont, ruleIds)
                elif gran == 'block':
                    try:
                        blocks = self.rucio.getBlocksInContainer(dataCont)
                        for block in blocks:
                            for rule in self.rucio.listDataRules(block, account=rucioAcct):
                                wflow['RulesToClean'][currPline].append(rule['id'])
                                msg = "Found %s block-level rule to be deleted for container %s"
                                self.logger.info(msg, rule['id'], dataCont)
                    except WMRucioDIDNotFoundException:
                        msg = "Container: %s not found in Rucio for workflow: %s."
                        self.logger.info(msg, dataCont, wflow['RequestName'])
        return wflow

    def cleanRucioRules(self, wflow):
        """
        Cleans all the Rules present in the field 'RulesToClean' in the MSRuleCleaner
        workflow representation. And fills the relevant Cleanup Status.
        Lifetime of rule will be updated to 0 rather than deleted, following
        issue: https://github.com/dmwm/CMSRucio/issues/921
        :param wflow:   A MSRuleCleaner workflow representation
        :return:        The workflow object
        """
        # NOTE: The function should be called independently and sequentially from
        #       The Input and the respective BlockLevel pipelines.

        # NOTE: The current functional pipeline is always the last one in the PlineMarkers list
        currPline = wflow['PlineMarkers'][-1]
        delResults = []
        if self.msConfig['enableRealMode']:
            for rule in wflow['RulesToClean'][currPline]:
                self.logger.info("%s: Updating lifetime=0 to ruleId: %s ", currPline, rule)
                delResult = self.rucio.updateRule(rule, {"lifetime": 0})
                delResults.append(delResult)
                if not delResult:
                    self.logger.warning("%s: Failed to update ruleId: %s ", currPline, rule)
        else:
            for rule in wflow['RulesToClean'][currPline]:
                delResults.append(True)
                self.logger.info("%s: DRY-RUN: Is about to update ruleId: %s ", currPline, rule)

        # Set the cleanup flag:
        wflow['CleanupStatus'][currPline] = all(delResults)
        return wflow

    def getRequestRecords(self, reqStatus):
        """
        Queries ReqMgr2 for requests in a given status.
        :param reqStatus: The status for the requests to be fetched from ReqMgr2
        :return requests: A dictionary with all the workflows in the given status
        """
        self.logger.info("Fetching requests in status: %s", reqStatus)
        result = self.reqmgr2.getRequestByStatus([reqStatus], detail=True)
        if not result:
            requests = {}
        else:
            requests = result[0]
        self.logger.info('  retrieved %s requests in status: %s', len(requests), reqStatus)
        return requests

    def alertStatusAdvanceExpired(self, wflow):
        """
        Send alert notifying that a workflow has spend too much time in a given status
        :param wflow: MSRuleCleanerWorkflow instance
        :additionalInfo: String with additional information.
        :return: none
        """
        lastTransitionTime = self._getLastStatusTransitionTime(wflow)
        alertName = "{}: Not archived workflow: {}".format(self.alertServiceName, wflow['RequestName'])
        alertSeverity = "medium"
        alertDescription = "Found a workflow not archived since {}.".format(time.ctime(lastTransitionTime))
        alertDescription += "\nWorkflow: \n{}".format(pformat({wfKey: wflow[wfKey] for wfKey in ['RequestName',
                                                                                                 'ParentageResolved',
                                                                                                 'TransferDone',
                                                                                                 'TransferTape',
                                                                                                 'TapeRulesStatus']}))
        alertDescription += "\nHas exceeded the Status Advance Timeout of: {} hours".format(self.msConfig['archiveAlarmHours'])
        alertDescription += " for its current status: {}.".format(wflow['RequestStatus'])
        if wflow['StatusAdvanceExpiredMsg']:
            alertDescription += "\nAdditional info: \n{}".format(wflow['StatusAdvanceExpiredMsg'])
        alertDescription += "\nActions need to be taken!"
        alertSummary = "[MSRuleCleaner] Found workflow: {} not archived since {}.".format(wflow['RequestName'], time.ctime(lastTransitionTime))

        # Check if alarms are enabled for this service
        # Alert expiration time defaults to 2 days
        if self.msConfig["sendNotification"]:
            tag = self.alertDestinationMap.get("alertStatusAdvanceExpired", "")
            self.sendAlert(alertName, alertSeverity, alertSummary, alertDescription,
                           service=self.alertServiceName, endSecs=self.alertExpiration, tag=tag)
        self.logger.critical(alertDescription)



from __future__ import (division, print_function)
from future import standard_library

from Utils.Utilities import decodeBytesToUnicode

standard_library.install_aliases()

import logging
from os import path
from urllib.parse import urlparse

from collections import defaultdict
from WMCore.Services.Service import Service
from WMCore.Services.TagCollector.XMLUtils import xml_parser


class TagCollector(Service):
    """
    Class which provides interface to CMS TagCollector web-service.
    Provides non-deprecated CMSSW releases in all their ScramArchs (not only prod)
    """

    def __init__(self, url=None, logger=None, configDict=None, **kwargs):
        """
        responseType will be either xml or json
        """
        defaultURL = "https://cmssdt.cern.ch/SDT/cgi-bin/ReleasesXML"
        url = url or defaultURL
        parsedUrl = urlparse(url)
        self.cFileUrlPath = parsedUrl.path.replace("/", "_")
        # all releases types and all their archs
        self.tcArgs = kwargs
        self.tcArgs.setdefault("anytype", 1)
        self.tcArgs.setdefault("anyarch", 1)

        configDict = configDict or {}
        configDict.setdefault('endpoint', url)
        configDict.setdefault("timeout", 300)
        configDict.setdefault('cacheduration', 1)
        configDict['logger'] = logger if logger else logging.getLogger()
        super(TagCollector, self).__init__(configDict)
        self['logger'].debug("Initializing TagCollector with url: %s", self['endpoint'])

        self.cvmfsReleasesMap = "/cvmfs/cms.cern.ch/releases.map"
        self.tmpReleasesXML = "/tmp/ReleasesXML"

    def parseCvmfsReleasesXML(self, releasesMap, releasesXML):
        """
        _parseCvmfsReleasesXML_
        
        Parses the ReleasesXML file from the releases.map in cvmfs
        """
        production = "type=Production;"
        announced = "state=Announced;"
        anyarch = False
        architecture = ""

        with open(releasesMap, "r", encoding="utf-8") as releasesFile:
            archs = {}
            for line in releasesFile:
                rels = []
                if not anyarch and 'prodarch=1;' not in line:
                    continue

                if production and production not in line:
                    continue

                if announced and announced not in line:
                    continue

                if architecture and architecture not in line:
                    continue

                data = {}
                for item in line.split(";"):
                    if "=" not in item:
                        continue
                    k, v = item.split("=")
                    data[k] = v

                if "architecture" in data and "label" in data and "type" in data and "state" in data:
                    if not anyarch and data["label"] in rels:
                        continue

                    rels.append(data["label"])
                    arch = data["architecture"]

                    if arch not in archs:
                        archs[arch] = []

                    extraTag = ""
                    if "default_micro_arch" in data:
                        extraTag = ' default_micro_arch="%s"' % data["default_micro_arch"]

                    data["extra_tag"] = extraTag
                    archs[arch].append(
                        """<project label="%(label)s" type="%(type)s" state="%(state)s"%(extra_tag)s/>"""
                        % data
                        )

        with open(releasesXML, "w", encoding="utf-8") as xml:
            xml.write("<projects>\n")
            for arch in archs:
                xml.write('  <architecture name="%s">\n' % arch)
                for rel in archs[arch]:
                    xml.write("    %s\n" % rel)
                xml.write("  </architecture>\n")
            xml.write("</projects>\n")

        return

    def _getResult(self, callname="", clearCache=False,
                   args=None, verb="GET", encoder=None, decoder=None,
                   contentType=None):
        """
        _getResult_

        retrieve JSON/XML formatted information given the service name and the
        argument dictionaries

        TODO: Probably want to move this up into Service
        """

        try:
            if not args:
                args = self.tcArgs

            cFile = '%s_%s' % (self.cFileUrlPath, callname.replace("/", "_"))
            # If no callname or url path, the base host is getting queried
            if cFile == '_':
                cFile = 'baseRequest'

            if clearCache:
                self.clearCache(cFile, args, verb)

            # Note cFile is just the base name pattern, args 
            # are also considered for the end filename in the method below
            f = self.refreshCache(cFile, callname, args, encoder=encoder, decoder=decodeBytesToUnicode,
                                verb=verb, contentType=contentType)
            result = f.read()
            f.close()

        except Exception:
            logging.error('Something went wrong accessing ReleasesXML from cmssdt, perhaps the service is temporarily down')
            logging.info('Checking if cvmfs is mounted')
            cvmfsMounted = path.ismount('/cvmfs')
            if cvmfsMounted:
                logging.info('cvmfs is mounted. Retrying to access ReleasesXML from cvmfs')
                try:
                    self.parseCvmfsReleasesXML(releasesMap=self.cvmfsReleasesMap, releasesXML=self.tmpReleasesXML)
                    with open(self.tmpReleasesXML, 'r', encoding='utf-8') as f:
                        result = f.read()
                    f.close()

                except Exception:
                    logging.error('Something went wrong parsing /cvmfs/cms.cern.ch/releases.map into XML format')
                    logging.exception('Unable to access ReleasesXML from cmssdt and cvmfs')
                    raise
            else:
                logging.info('cvmfs is not mounted, nothing else to do')
                raise

        # overhead from REST model which returns results as strings or None
        # therefore they can be encoded by JSON to None, etc.
        if result == 'None':
            return
        if result and decoder:
            result = decoder(result)
        return result

    def data(self):
        "Fetch data from tag collector or local cache"
        data = self._getResult()
        pkey = 'architecture'
        for row in xml_parser(data, pkey):
            yield row[pkey]

    
    def releases(self, arch=None):
        "Yield CMS releases known in tag collector"
        arr = []
        for row in self.data():
            if arch:
                if arch == row['name']:
                    for item in row['project']:
                        arr.append(item['label'])
            else:
                for item in row['project']:
                    arr.append(item['label'])
        return list(set(arr))

    def architectures(self):
        "Yield CMS architectures known in tag collector"
        arr = []
        for row in self.data():
            arr.append(row['name'])
        return list(set(arr))

    def releases_by_architecture(self):
        "returns CMS architectures and realease in dictionary format"
        arch_dict = defaultdict(list)
        for row in self.data():
            releases = set()
            for item in row['project']:
                releases.add(item['label'])
            arch_dict[row['name']].extend(list(releases))
        return dict(arch_dict)


    def defaultMicroArchVersionNumberByRelease(self, default_microarch=0):
        """
        Yield default microarchitecture by CMS release known in tag collector.

        :param default_microarch: int, default microar when not found in TagCollector XML file.
        :return: dictionary ("release": default_microarch (or default_microarch if not found)).
        """
        rel_microarchs = {}
        for row in self.data():
            for item in row['project']:
                microarch = item.get('default_micro_arch', default_microarch)
                if isinstance(microarch, str):
                    # format of this value is "arch-vN"
                    microarch = int((microarch.split("-")[-1]).lstrip("v"))
                rel_microarchs[item['label']] = microarch
        return rel_microarchs

    def getGreaterMicroarchVersionNumber(self, releases=None, default_microarch=0, rel_microarchs=None):
        """
        Return the greated default microarchitecture, given a list of releases.

        :param releases: str, comma-separated list of releases.
        :param default_microarch: int, default microar when not found in TagCollector XML file.
        :return: int, greater default microarchitecture version number.
        """
        if releases is None:
            return default_microarch

        if rel_microarchs is None:
            rel_microarchs = self.defaultMicroArchVersionNumberByRelease()

        releases = releases.split(",")
        microarch = default_microarch
        for r in releases:
            m = rel_microarchs.get(r, default_microarch)
            if m > microarch:
                microarch = m

        return microarch



#!/usr/bin/env python
"""
_Transaction_

A simple wrapper around DBInterface to make working with transactions simpler

On MySQL transactions only work for innodb tables.
"""



import logging
import time

from WMCore.DataStructs.WMObject import WMObject
from WMCore.WMException import WMException
from WMCore.WMExceptions import WMEXCEPTION

class Transaction(WMObject):
    dbi = None

    def __init__(self, dbinterface = None):
        """
        Get the connection from the DBInterface and open a new transaction on it
        """
        self.dbi = dbinterface
        self.conn = None
        self.transaction = None

    def begin(self):
        if self.conn == None:
            self.conn = self.dbi.connection()
        if self.conn.closed:
            self.conn = self.dbi.connection()

        if self.transaction == None:
            self.transaction = self.conn.begin()

        return

    def processData(self, sql, binds={}):
        """
        Propagates the request to the proper dbcore backend,
        and performs checks for lost (or closed) connection.
        """
        result = self.dbi.processData(sql, binds, conn = self.conn,
                                      transaction = True)
        return result

    def commit(self):
        """
        Commit the transaction and return the connection to the pool
        """
        if not self.transaction == None:
            self.transaction.commit()

        if not self.conn == None:
            self.conn.close()
        self.conn = None
        self.transaction = None

    def rollback(self):
        """
        To be called if there is an exception and you want to roll back the
        transaction and return the connection to the pool
        """
        if self.transaction:
            self.transaction.rollback()

        if self.conn:
            self.conn.close()

        self.conn = None
        self.transaction = None
        return

    def rollbackForError(self):
        """
        This is called when handling a major exception.  This is because sometimes
        you can end up in a situation where the transaction appears open, but is not.  In
        this case, calling a rollback on the transaction will cause an exception, which
        then destroys all logging and shutdown of the actual code.

        Use only in components.
        """

        try:
            self.rollback()
        except:
            pass
        return



"""
_DiscreteSummaryHistogram_

Discrete histogram module, to be used by the TaskArchiver
to store histograms in the summary where the data is categorical.

Created on Nov 20, 2012

@author: dballest
"""

from WMCore.DataStructs.MathStructs.SummaryHistogram import SummaryHistogram
from WMCore.Algorithms.MathAlgos import getAverageStdDev

class DiscreteSummaryHistogram(SummaryHistogram):
    """
    _DiscreteSummaryHistogram_

    A histogram where the data is organized by
    a finite number of categories, it can have
    many values for each category.
    """


    def __init__(self, title, xLabel):
        """
        __init__

        Initialize a simpler histogram that only stores
        the histogram. Everything else is calculated when the JSON is requested.
        """
        # Initialize the parent object
        SummaryHistogram.__init__(self, title, xLabel)

        # Indicate this is a discrete histogram
        self.continuous = False

        # Add data only used in the discrete version
        self.yLabels    = set()

        # Override initialization of some attributes
        self.average = {}
        self.stdDev  = {}

        return

    def addPoint(self, xValue, yLabel):
        """
        _addPoint_

        Add point to discrete histogram,
        x value is a category and therefore not rounded.
        There can be many yLabel and standard deviations are
        not calculated online. Histograms are always stored.
        """
        if xValue not in self.data:
            # Record the category
            self.data[xValue] = {}
            for label in self.yLabels:
                self.data[xValue][label] = 0

        if yLabel not in self.yLabels:
            # Record the label
            self.yLabels.add(yLabel)
            self.average[yLabel] = 0.0
            self.stdDev[yLabel]  = 0.0
            for category in self.data:
                self.data[category][yLabel] = 0

        self.data[xValue][yLabel] += 1

        return

    def __add__(self, other):
        #TODO: For HG1302, support multiple agents properly in the workload summary
        raise NotImplementedError

    def toJSON(self):
        """
        _toJSON_

        Calculate average and standard deviation, store it
        and call the parent class toJSON method
        """

        for yLabel in self.yLabels:
            numList = []
            for xValue in self.data:
                numList.append(self.data[xValue][yLabel])
            (self.average[yLabel], self.stdDev[yLabel]) = getAverageStdDev(numList)

        return SummaryHistogram.toJSON(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
File       : ProcFS.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: This module provides function which rely on Linux proc fs
"""
# system modules
import os

# 3rd party modules
try:
    import psutil
    usePsutil = True
except ImportError:
    usePsutil = False


def processStatus(pid, method=None):
    """
    Return status of the given process PID along with its threads using psutil.
    :param pid: Process ID to inspect
    :param method: define which method to use procfs or psutil
    :return: list of dictionaries of `{process, pid, status}` data-structure
    """
    if method == 'procfs' or usePsutil == False:
        return processStatusViaProcFS(pid)
    return processStatusViaPsutil(pid)


def processStatusViaPsutil(pid):
    """
    Return status of the given process PID along with its threads using psutil.
    :param pid: Process ID to inspect
    :return: list of dictionaries with {process, pid, status, type}
    """
    statusList = []

    try:
        proc = psutil.Process(int(pid))
        processName = proc.name()
        processStatus = proc.status()
        statusList.append({
            "process": processName,
            "pid": str(pid),
            "status": processStatus,
            "type": "process"
        })

        # Iterate over threads
        threads = proc.threads()
        for thread in threads:
            if str(pid) == str(thread.id):
                continue
            threadId = thread.id
            # psutil doesn't provide thread name or detailed state, so we approximate
            statusList.append({
                "process": f"{processName}-thread",
                "pid": str(threadId),
                "status": "running",  # threads are assumed running if present
                "type": "thread"
            })

    except psutil.NoSuchProcess:
        return [{"error": f"Process {pid} not found"}]
    except Exception as e:
        return [{"error": str(e)}]

    return statusList


def processStatusViaProcFS(pid):
    """
    Return status of given process PID along with its threads via /proc FS look-up (available on all Linux OSes)
    :return: list of dictionaries of `{process, pid, status}` data-structure
    """
    statusList = []
    procPath = f"/proc/{pid}"

    if not os.path.exists(procPath):
        return [{"error": f"Process {pid} not found"}]

    try:
        with open(f"{procPath}/status", encoding='utf-8') as f:
            processName = ""
            processState = ""
            for line in f:
                if line.startswith("Name:"):
                    processName = line.split(":")[1].strip()
                elif line.startswith("State:"):
                    processState = line.split(":")[1].strip()
            statusList.append({"process": processName, "pid": str(pid), "status": processState, "type": "process"})

        taskPath = f"{procPath}/task"
        if os.path.exists(taskPath):
            for threadId in os.listdir(taskPath):
                if str(threadId) == str(pid):
                    continue
                with open(f"{taskPath}/{threadId}/status", encoding='utf-8') as f:
                    threadName = ""
                    threadState = ""
                    for line in f:
                        if line.startswith("Name:"):
                            threadName = line.split(":")[1].strip()
                        elif line.startswith("State:"):
                            threadState = line.split(":")[1].strip()
                    statusList.append({"process": threadName, "pid": threadId, "status": threadState, "type": "thread"})

    except Exception as e:
        return [{"error": str(e)}]

    return statusList



"""
A basic action is a thing that will run a SQL statement

A more complex one would be something that ran multiple SQL
objects to produce a single output.
"""

from builtins import object

class DAOFactory(object):
    def __init__(self, package='WMCore', logger=None, dbinterface=None, owner=""):
        self.package = package
        self.logger = logger
        self.dbinterface = dbinterface
        self.owner = owner
        #self.logger.debug("Instantiating DAOFactory for %s package" % self.package)
        from WMCore.Database.Dialects import MySQLDialect
        from WMCore.Database.Dialects import OracleDialect
        self.dialects = {"Oracle" : OracleDialect,
                    "MySQL" : MySQLDialect,}

    def __call__(self, classname):
        """
        Somewhat fugly method to load generic SQL classes...
        """
        if not isinstance(self.dbinterface, str):

            dia = self.dbinterface.engine.dialect
            #TODO: Make good
            dialect = None
            for i in self.dialects:
                if isinstance(dia, self.dialects[i]):
                    dialect = i
            if not dialect:
                raise TypeError("unknown connection type: %s" % dia)
        else:
            dialect = 'CouchDB'


        module = "%s.%s.%s" % (self.package, dialect, classname)
        #self.logger.debug("importing %s, %s" % (module, classname))
        module = __import__(module, globals(), locals(), [classname])#, -1)
        instance = getattr(module, classname.split('.')[-1])
        if self.owner:
            return instance(self.logger, self.dbinterface, self.owner)
        else:
            return instance(self.logger, self.dbinterface)



#!/usr/bin/env python
# encoding: utf-8
"""
User.py

Created by Dave Evans on 2010-07-14.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from WMCore.GroupUser.CouchObject import CouchObject
import WMCore.GroupUser.Decorators as Decorators
from WMCore.GroupUser.Group import Group



class User(CouchObject):
    """
    _User_


    """
    def __init__(self, **options):
        CouchObject.__init__(self)
        self.cdb_document_data = "user"
        self.setdefault('name', None)
        self.setdefault('proxy', {})
        self.couch_url = None
        self.couch_db = None
        self.group = None
        self.update(options)

    document_id = property(lambda x : "user-%s" % x['name'] )
    name = property(lambda x: x['name'])

    def setGroup(self, groupInstance):
        """
        _setGroup_

        Set the group attribute of this object
        """
        self.group = groupInstance
        self['group'] = groupInstance['name']
        if groupInstance.connected:
            # if the group already has the couch instance in it, borrow those settings
            self.setCouch(groupInstance.cdb_url, groupInstance.cdb_database)

    @Decorators.requireConnection
    @Decorators.requireGroup
    def ownThis(self, document):
        """
        _ownThis_

        Given a Couch Document, insert the details of this owner in a standard
        way so that the document can be found using the standard group/user views
        Note: if doc doesnt have both _id and _rev keys this change wont stick
        """
        document['owner'] = {}
        document['owner']['user'] = self['name']
        document['owner']['group'] = self.group['name']

        retval = self.couch.commitOne(document)
        document["_id"] = retval[0]["id"]
        document["_rev"] = retval[0]["rev"]
        return

    @Decorators.requireConnection
    @Decorators.requireGroup
    def create(self):
        """
        _create_

        Overide the base class create to make sure the group exists when adding the user
        (Note: Overriding wipes out decorators...)
        """
        if not self.couch.documentExists(self.group.document_id):
            # no group => create group
            self.group.create()
        # call base create to make user
        CouchObject.create(self)

#ToDo: Override drop to drop all documents owned by the owner...

def makeUser(groupname, username, couchUrl = None, couchDatabase = None):
    """
    _makeUser_

    factory like util that creates a user and group object

    """
    group = Group(name = groupname)
    if couchUrl != None:
        group.setCouch(couchUrl, couchDatabase)
        group.connect()
    user  = User(name = username)
    user.setGroup(group)
    return user



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : simulateReplicationFailure.py
author     : valentin kuznetsov <vkuznet AT gmail DOT com>
Description: helper script to simulate replication failure
"""

import os
import sys
import json
import tempfile
import time
import requests

# 3rd party modules
from Utils.CertTools import ckey, cert, caBundle


def wipeDb(couchUrl, dbName, auth=None, certTuple=None, caFileName=None):
    """
    wipe out couch database
    :param couchUrl: couch url string
    :param dbName: database name
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    deleteResp = requests.delete(f"{couchUrl}/{dbName}", auth=auth, cert=certTuple, verify=caFileName)

    if deleteResp.status_code not in (200, 202, 404):
        raise Exception(f"Failed to delete database: {deleteResp.text}")
    print(f"Deleted database '{dbName}' (status {deleteResp.status_code})")

def createDb(couchUrl, dbName, auth, certTuple=None, caFileName=None):
    """
    creates new couchdb database
    :param couchUrl: couch url string
    :param dbName: database name
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    r = requests.put(f"{couchUrl}/{dbName}", auth=auth, cert=certTuple, verify=caFileName)
    if r.status_code not in (201, 412):  # 412 = already exists
        raise Exception(f"Failed to create DB {couchUrl}/{dbName}: {r.text} status code {r.status_code}")

def setupValidator(couchUrl, targetDb, auth, certTuple=None, caFileName=None):
    """
    setup validator in couch database
    :param couchUrl: couch url string
    :param targetDb: database name
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    designDoc = {
        "_id": "_design/rejectBadDocs",
        "validate_doc_update": (
            "function(newDoc, oldDoc, userCtx) { "
            "if (newDoc.blockMe === true) { throw({forbidden: 'Blocked by validator'}); } "
            "}"
        )
    }
    headers = {'Content-Type': 'application/json'}

    r = requests.put(f"{couchUrl}/{targetDb}/{designDoc['_id']}",
                     data=json.dumps(designDoc), headers=headers, auth=auth, cert=certTuple, verify=caFileName)
    if r.status_code not in (201, 202):
        raise Exception(f"Failed to upload validator: {r.text}")

def insertValidDocs(couchUrl, sourceDb, auth, certTuple=None, caFileName=None):
    """
    insert set of valid documents into couch database
    wipe out couch database
    :param couchUrl: couch url string
    :param sourceDb: database name
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    validDocIds = []
    headers = {'Content-Type': 'application/json'}
    for i in range(3):
        doc = {"_id": f"validDoc{i}", "msg": f"This is valid doc #{i}"}
        r = requests.put(f"{couchUrl}/{sourceDb}/{doc['_id']}",
                         data=json.dumps(doc), headers=headers, auth=auth, cert=certTuple, verify=caFileName)
        if r.status_code in (201, 202):
            validDocIds.append(doc['_id'])
        else:
            raise Exception(f"Failed to insert valid doc: {r.text}")
    return validDocIds

def insertInvalidDoc(couchUrl, sourceDb, auth, certTuple=None, caFileName=None):
    """
    insert invalid document into couch database
    :param couchUrl: couch url string
    :param sourceDb: database name
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    doc = {"_id": "badDoc", "blockMe": True, "msg": "This should fail replication"}
    headers = {'Content-Type': 'application/json'}
    r = requests.put(f"{couchUrl}/{sourceDb}/{doc['_id']}",
                     data=json.dumps(doc), headers=headers, auth=auth, cert=certTuple, verify=caFileName)
    if r.status_code not in (201, 202):
        raise Exception(f"Failed to insert bad doc: {r.text}")
    return doc['_id']

def deleteReplicationDoc(sourceUrl, replicationId, auth, certTuple=None, caFileName=None):
    """
    delete replication document
    :param sourceUrl: source url string
    :param replicationId: replication id string
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    docUrl = f"{sourceUrl}/_replicator/{replicationId}"
    r = requests.get(docUrl, auth=auth, cert=certTuple, verify=caFileName)

    if r.status_code == 200:
        rev = r.json()['_rev']
        delResp = requests.delete(f"{docUrl}?rev={rev}", auth=auth, cert=certTuple, verify=caFileName)
        if delResp.status_code not in (200, 202):
            raise Exception(f"Failed to delete replication doc: {delResp.text}")
        print(f"Deleted existing replication doc '{replicationId}'")
    elif r.status_code != 404:
        raise Exception(f"Failed to check replication doc: {r.text}")

def setupReplication(sourceUrl, sourceDb, targetUrl, targetDb, auth, replicationId, certTuple=None, caFileName=None):
    """
    setup replication between source and target databases
    :param sourceUrl: source couchdb url string
    :param sourceDb: source database name
    :param targetUrl: target couchdb url string
    :param targetDb: target database name
    :param auth: authentication tuple
    :param replicationId: replication id string
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    replicationDoc = {
        "_id": replicationId,
        "source": sourceDb,
        "target": f"{targetUrl}/{targetDb}",
        "create_target": False,
        "continuous": True,
        "user_ctx": {"name": "admin", "roles": ["_admin"]}
    }
    headers = {'Content-Type': 'application/json'}
    r = requests.put(f"{sourceUrl}/_replicator/{replicationId}",
                     data=json.dumps(replicationDoc), headers=headers, auth=auth, cert=certTuple, verify=caFileName)
    if r.status_code not in (201, 202):
        raise Exception(f"Failed to create replication: {r.text}")
    print(f"Replication '{replicationId}' set up successfully with status {r.status_code}")

def checkReplicationStatus(couchUrl, replicationId, auth, certTuple=None, caFileName=None):
    """
    checks replication status for given replication id
    :param couchUrl: couch url string
    :param replicationId: replication id string
    :param auth: authentication tuple
    :param certTuple: cert and ckey tuple
    :param caFileName: CA file name
    :return: nothing
    """
    for _ in range(10):
        time.sleep(2)
        r = requests.get(f"{couchUrl}/_replicator/{replicationId}", auth=auth, cert=certTuple, verify=caFileName)
        doc = r.json()
        state = doc.get("_replication_state", "unknown")
        reason = doc.get("_replication_state_reason", "")
        print(f"Replication state: {state}")
        if reason:
            print(f"Reason: {reason}")
        if state in ("error", "completed"):
            return state, reason
    return "unknown", ""

def main():
    """
    main function for testing purposes
    """
    if os.getenv('COUCH_SOURCE_URL') == '':
        print('ERROR: no COUCH_SOURCE_URL found')
        sys.exit(1)
    if os.getenv('COUCH_TARGET_URL') == '':
        print('ERROR: no COUCH_TARGET_URL found')
        sys.exit(1)
    if os.getenv('COUCH_CREDENTIALS') == '':
        print('ERROR: no COUCH_CREDENTIALS found, should be username:password')
        sys.exit(1)

    certTuple = (cert(), ckey())
    capath = os.environ.get("X509_CERT_DIR", '/etc/grid-security/certificates')
    cacerts = caBundle(capath)

    with tempfile.NamedTemporaryFile(mode='w+', suffix=".pem", delete=True) as tfile:
        tfile.write(cacerts)
        tfile.flush()
        caFileName = tfile.name

        sourceUrl = os.getenv('COUCH_SOURCE_URL', '')
        targetUrl = os.getenv('COUCH_TARGET_URL', '')
        authEnv = os.getenv('COUCH_CREDENTIALS', '')
        auth = tuple(authEnv.split(':')) if ':' in authEnv else None
        sourceDb = 'test_db'
        targetDb = 'test_db'
        replicationId = 'replicationFailTest'

        print(f"INFO: wipe out {sourceUrl}/{sourceDb}")
        wipeDb(sourceUrl, sourceDb, auth, certTuple, caFileName)
        print(f"INFO: wipe out {targetUrl}/{targetDb}")
        wipeDb(targetUrl, targetDb, auth, certTuple, caFileName)

        print("INFO: Setting up test environment...")
        createDb(sourceUrl, sourceDb, auth, certTuple, caFileName)
        createDb(targetUrl, targetDb, auth, certTuple, caFileName)
        setupValidator(targetUrl, targetDb, auth, certTuple, caFileName)

        print("INFO: Inserting valid documents...")
        validDocIds = insertValidDocs(sourceUrl, sourceDb, auth, certTuple, caFileName)

        print("INFO: Inserting a document that will be rejected by the target...")
        badDocId = insertInvalidDoc(sourceUrl, sourceDb, auth, certTuple, caFileName)

        print("INFO: Starting replication...")
        deleteReplicationDoc(sourceUrl, replicationId, auth, certTuple, caFileName)
        setupReplication(sourceUrl, sourceDb, targetUrl, targetDb, auth, replicationId, certTuple, caFileName)

        print("INFO: Waiting for replication to complete or fail...")
        state, reason = checkReplicationStatus(sourceUrl, replicationId, auth, certTuple, caFileName)

        print("\nINFO: Successfully inserted document IDs:")
        for docId in validDocIds:
            print(f" - {docId}")

        print(f"\nWARNING: Problematic document ID (should fail replication): {badDocId}")

        print(f"\nINFO: Replication final state: {state}")
        if reason:
            print(f"ERROR: Replication failure reason: {reason}")
        else:
            print("INFO: No error reported.")

if __name__ == "__main__":
    main()




"""
_SummaryHistogram_

Histogram module, to be used by the TaskArchiver
to store histograms in the summary.

Created on Nov 16, 2012

@author: dballest
"""
from builtins import str
from WMCore.DataStructs.WMObject import WMObject

class SummaryHistogram(WMObject):
    """
    _SummaryHistogram_

    Histogram object, provides familiar CRUD methods
    which take care of most of the statistical
    calculations when adding points, this object
    can also be converted into a dictionary
    for JSON documents. It knows how to combine
    with other histograms and create itself from
    a dictionary provided it has matching structure.

    This is an interface, the real work is done
    by the ContinuousSummaryHistogram and
    DiscreteSummaryHistogram objects
    """

    def __init__(self, title = None, xLabel = None):
        """
        __init__
        Initialize the elements in the object.
        """

        # Meta-information about the histogram, it can be changed at any point
        self.title      = title
        self.xLabel     = xLabel

        # These shouldn't be touched from anything outside the SummaryHistogram object and children classes
        self.continuous         = None
        self.jsonInternal       = None
        self.data               = {}
        self.average            = None
        self.stdDev             = None

        return

    def setTitle(self, newTitle):
        """
        _setTitle_

        Set the title
        """
        self.title = newTitle
        return

    def setHorizontalLabel(self, xLabel):
        """
        _setHorizontalLabel_

        Set the label on the x axis
        """
        self.xLabel = xLabel
        return

    def addPoint(self, xValue, yLabel):
        """
        _addPoint_

        Add a point to the histogram data, a histogram
        can have many types of y values for the same x if
        x is continuous otherwise it is only one yLabel.
        They should be in a similar scale for best results.
        """
        raise NotImplementedError("SummaryHistogram objects can't be used, use either the continuous or discrete implementation")

    def toJSON(self):
        """
        _toJSON_

        Return a dictionary which is compatible
        with a JSON object
        """
        if self.continuous is None:
            raise TypeError("toJSON can't be called on a bare SummaryHistogram object")

        # Get what the children classes did
        jsonDict = {}
        jsonDict['internalData'] = self.jsonInternal or {}

        # Add the common things
        jsonDict['title']      = self.title
        jsonDict['xLabel']     = self.xLabel
        jsonDict['continuous'] = self.continuous
        jsonDict['data']       = self.data
        jsonDict['stdDev']     = self.stdDev
        jsonDict['average']    = self.average

        return jsonDict

    def __add__(self, other):
        """
        __add__

        Add two histograms, combine statistics.
        """
        raise NotImplementedError("SummaryHistogram objects can't be used, use either the continuous or discrete implementation")

    def __str__(self):
        """
        __str__

        Return the str object of the JSON
        """
        return str(self.toJSON())



"""
_PyCondorAPI_

Class used to interact with Condor daemons on the agent
"""

from __future__ import print_function, division

from builtins import str, object
from past.builtins import basestring
import logging
try:
    # This module has dependency with python binding for condor package (condor)
    import htcondor2 as htcondor
except ImportError:
    pass


class PyCondorAPI(object):
    """
    Some APIs to interact with HTCondor via the HTCondor python bindings.
    """
    def __init__(self):
        self.schedd = htcondor.Schedd()
        self.coll = htcondor.Collector()

    def recreateSchedd(self):
        """
        In case our current schedd object is in a "strange" state, we
        better recreate it.
        """
        logging.warning("Recreating Schedd instance due to query error...")
        self.schedd = htcondor.Schedd()

    def getCondorJobsSummary(self):
        """
        Retrieves a job summary from the HTCondor Schedd object
        :return: a list of classads representing the matching jobs, or None if failed
        """
        jobs = None  # return None to signalize the query failed
        queryOpts = htcondor.QueryOpts.SummaryOnly
        try:
            return self.schedd.query(opts=queryOpts)
        except Exception:
            self.recreateSchedd()

        try:
            jobs = self.schedd.query(opts=queryOpts)
        except Exception as ex:
            logging.exception("Failed to fetch summary of jobs from Condor Schedd. Error: %s", str(ex))
        return jobs

    def getCondorJobs(self, constraint='true', attrList=None, limit=-1, opts="Default"):
        """
        Given a job/schedd constraint, return a list of job classad.
        :param constraint: the query constraint (str or ExprTree). Defaults to 'true'
        :param attrList: a list of attribute strings to be returned in the call.
                         It defaults to all attributes.
        :param limit: a limit on the number of matches to return. Defaults to -1 (all)
        :param opts: string with additional flags for the query. Defaults to Default.
            https://htcondor.readthedocs.io/en/v8_9_7/apis/python-bindings/api/htcondor.html#htcondor.QueryOpts
        :return: returns an iterator to the job classads
        """
        attrList = attrList or []
        # if option parameter is invalid, default it to the standard behavior
        opts = getattr(htcondor.htcondor.QueryOpts, opts, "Default")
        msg = "Querying condor schedd with params: constraint=%s, attrList=%s, limit=%s, opts=%s"
        logging.info(msg, constraint, attrList, limit, opts)
        try:
            return self.schedd.query(constraint, attrList, limit, opts=opts)
        except Exception:
            self.recreateSchedd()

        # if we hit another exception, let it be raised up in the chain
        return self.schedd.query(constraint, attrList, limit, opts=opts)

    def editCondorJobs(self, job_spec, attr, value):
        """
        _editCondorJobs_

        Edit a set of condor jobs given an attribute and value
        job_spec can be a list of job IDs or a string specifying a constraint
        """
        success = False
        try:
            self.schedd.edit(job_spec, attr, value)
            success = True
        except Exception as ex:
            # edit doesn't distinguish between an error and not matching any jobs
            # check for this message and assume it just didn't match any jobs
            if isinstance(ex, RuntimeError) and str(ex) == "Unable to edit jobs matching constraint":
                success = True
                msg = "Condor constraint did not match any jobs. "
                msg += "Message from schedd: %s" % str(ex)
                logging.info(msg)
            else:
                msg = "Condor failed to edit the jobs. "
                msg += "Error message: %s" % str(ex)
                logging.exception(msg)

        return success

    def isScheddOverloaded(self):
        """
        check whether job limit is reached in local schedd.
        Condition is check by following logic.
        ( ShadowsRunning > 9.700000000000000E-01 * MAX_RUNNING_JOBS) )
        || ( RecentDaemonCoreDutyCycle > 9.800000000000000E-01 )
        """
        try:
            scheddAd = self.coll.locate(htcondor.DaemonTypes.Schedd)
            q = self.coll.query(htcondor.AdTypes.Schedd, 'Name == "%s"' % scheddAd['Name'],
                                projection=['CurbMatchmaking'])[0]
            isOverloaded = q['CurbMatchmaking'].eval(q)
            return isOverloaded
        except Exception:
            # if there is an error, try to recreate the collector instance
            logging.info("Recreating Collector instance due to query error...")
            self.coll = htcondor.Collector()
        try:
            scheddAd = self.coll.locate(htcondor.DaemonTypes.Schedd)
            q = self.coll.query(htcondor.AdTypes.Schedd, 'Name == "%s"' % scheddAd['Name'],
                                projection=['CurbMatchmaking'])[0]
            isOverloaded = q['CurbMatchmaking'].eval(q)
        except Exception as ex:
            msg = "Condor failed to fetch schedd attributes."
            msg += "Error message: %s" % str(ex)
            logging.exception(msg)
            # since it failed, assume it's overloaded
            isOverloaded = True

        return isOverloaded


def getScheddParamValue(param):
    """
    _getScheddParamValue_

    Given a schedd parameter, retrieve it's value with htcondor, e.g.:
    MAX_JOBS_RUNNING, MAX_JOBS_PER_OWNER, etc
    """
    paramResult = None
    if not isinstance(param, basestring):
        logging.error("Parameter %s must be string type", param)
        return paramResult

    try:
        paramResult = htcondor.param[param]
    except Exception as ex:
        msg = "Condor failed to fetch schedd parameter: %s" % param
        msg += "Error message: %s" % str(ex)
        logging.exception(msg)

    return paramResult
